/*> bp.c
 * Breakpoint manipulation and control routines.
 *
 * $Revision: 0.1 $
 *   $Author: jsmith $
 *     $Date: 941030 $
 *
 * Copyright (c) 1994 VLSI Technology Inc. All rights reserved.
 *
 * "@(#)bp.c	1.2	96/06/07 VLSI"
 */
/*---------------------------------------------------------------------------*/

#include    "global.h"
#include    "bp.h"
#include    "driver.h"

/*---------------------------------------------------------------------------*/
/* BP_init
 * =======
 *
 * ENTER:   void
 *
 * EXIT:    void
 *
 * This must be called prior to using break points.
 * This routine clears ALL the breakpoints.
 */
void BP_init(void)
{
 int  j ;
 int *addr ;
 int *inst ;

#if defined(VY86LX650)
 if (PONimage != 0) /* is this a Power On Reset? */
#endif
  {
   addr = (int *)(&BP_addr) ;
   inst = (int *)(&BP_inst) ;

   for (j=0; (j < NumBP); j++)
    {
     addr[j] = 0 ; /* breakpoint is clear if both addr and inst are ZERO */
     inst[j] = 0 ;
    }
  }

 return ;
}

/*---------------------------------------------------------------------------*/
/* BP_Set
 * ======
 *
 * ENTER:   ofs = offset to BP_addr and BP_inst
 *          adr = Addr at which to set breakpoint
 *
 * EXIT:        void
 *
 * NOTE: This code assumes that break point #1 starts at 0x24 and that
 *       each breakpoint (up to NumBP) is in the next consecutive loc.
 *       If this changes, the MOV pc,#(0x24 + offset) must be recoded.
 */
void BP_Set(int ofs,int adr)
{
 int *addr ;
 int *inst ;

 addr = (int *)((int)&BP_addr) ; /* address holding table */
 inst = (int *)((int)&BP_inst) ; /* instruction holding table */
 
 addr[ofs] = adr ;   /* save address of breakpoint */
 addr = (int *)adr ; /* point at instruction at breakpoint */
 inst[ofs] = *addr ; /* save instruction at breakpoint */

#if defined(RUBY2) && defined(HWBREAKPOINT)
 *((unsigned char *)(IOBase + DEBUG)) = (unsigned char)DEBUG_MATCH_OFF ;
 *((int *)(IOBase + ADDR_MATCH)) = adr ; /* address we want to match */
 *((unsigned char *)(IOBase + DEBUG)) = (unsigned char)DEBUG_MATCH_READS ; /* and enable address match abort */
#else /* software breakpoint support */
 /* replace the current instruction with a branch into the BP table */
 *addr = (0xE3A0F024 + (ofs * 4)) ; /* make a "MOV pc,#0x24+offset" instruction */
#endif /* RUBY2 and HWBREAKPOINT */

 return ;
}

/*---------------------------------------------------------------------------*/
/* BP_Clear
 * ========
 *
 * ENTER:       ofs = offset to BP_addr and BP_inst
 *
 * EXIT:        void
 *
 * This routine assumes that a breakpoint is to be cleared. It first restores
 * the instruction to the address the breakpoint was placed, and then clears
 * away the break point. A breakpoint is clear if both inst and addr are ZERO.
 */
void BP_Clear(int ofs)
{
 int *addr ;
 int *inst ;

#if defined(RUBY2) && defined(HWBREAKPOINT)
 /* We know we only have a single breakpoint */
 *((unsigned char *)(IOBase + DEBUG)) = (unsigned char)DEBUG_MATCH_OFF ;
 *((int *)(IOBase + ADDR_MATCH)) = 0x00000000 ; /* no address */

 addr = (int *)((int)&BP_addr) ; /* address holding table */
 inst = (int *)((int)&BP_inst) ; /* instruction holding table */
#else /* software breakpoint support */
 int *adr ;

 addr = (int *)((int)&BP_addr) ; /* address holding table */
 inst = (int *)((int)&BP_inst) ; /* instruction holding table */

 adr = (int *)addr[ofs] ; /* place address of instruction in adr */
 *adr = inst[ofs] ;       /* restore inst to proper location */
#endif /* RUBY2 and HWBREAKPOINT */
    
 addr[ofs] = 0 ; /* breakpoint is clear if both addr and inst are ZERO */
 inst[ofs] = 0 ;

 return ;
}

/*---------------------------------------------------------------------------*/
/* BP_Exist
 * ========
 *
 * ENTER:   adr = addr of a breakpoint
 *
 * EXIT:    TRUE (-1 or non-ZERO) if breakpoint present
 *          (or) FALSE (ZERO) if breakpoint does NOT exist
 */
int BP_Exist(int adr)
{
 int  result ;
 int  j ;
 int *addr ;

#if defined(RUBY2) && defined(HWBREAKPOINT)
 /* We could check to see if the address matches the value in the
  * address match register. However, for the moment we simply
  * perform the same table check as the software breakpoint code.
  */
#endif /* RUBY2 and HWBREAKPOINT */
 
 addr = (int *)((int)&BP_addr) ; /* address holding table */

 result = FALSE ; /* default assumes "adr" is NOT a breakpoint */
 j = 0 ;
 while (!result && (j < NumBP)) /* loop through breakpoint table */
  {
   if(addr[j++] == adr)
    result = TRUE ; /* "adr" is a breakpoint */
  }

 return(result) ;
}

/*---------------------------------------------------------------------------*/
/* BP_Offset
 * =========
 *
 * ENTER:   adr = addr of a breakpoint (or ZERO for 1st available)
 *
 * EXIT:    offset into BP_addr and BP_inst for breakpoint
 *          (or) -1 if breakpoint does NOT exist
 *
 * NOTE: This code assumes that BP_Exist has been called to confirm that a
 *       breakpoint does exist at addr passed in r0
 */
int BP_Offset(int adr)
{
#if defined(RUBY2) && defined(HWBREAKPOINT)
 /* Since this function definition states that BP_Exist *MUST* have
  * been called we should just be able to return zero, since we only
  * have a single breakpoint.
  */
 return(0) ;

 /* If we do not explicitly return zero, and use the software
  * checking loop below, then armsd will not run, since it will run out
  * of break-points when single-stepping. NOTE: This means that
  * single-stepping and break-points will *NOT* work when using the
  * hardware breakpoint feature of the RUBY2 device.
  */
#else /* software breakpoint world */
 int  result ;
 int  j ;
 int *addr ;

 addr = (int *)((int)&BP_addr) ; /* address holding table */
 result = -1 ; /* default assumes "adr" is NOT a breakpoint */
 j = 0 ;
 while ((result == -1) && (j < NumBP))
  {
   if (addr[j] == adr)
    result = j ; /* "adr" is a breakpoint, return offset into array */
   j++ ;
  }

 return(result) ;
#endif /* RUBY2 and HWBREAKPOINT */
}

/*---------------------------------------------------------------------------*/
/*> EOF bp.c <*/
