/*> driver.c <*/
/*---------------------------------------------------------------------------*/
/* This file provides the main VY86PID target specific I/O support.
 * The VY86PID board uses a 16C551 UART as the serial device.
 *
 * $Revision: 0.1 $
 *   $Author: jsmith $
 *     $Date: 941031 $
 *
 * Copyright (c) 1994, VLSI Technology Inc. All Rights Reserved.
 *
 * "@(#)driver.c	1.5	96/06/07 VLSI"
 */
/*---------------------------------------------------------------------------*/

#if !defined(VY86PID)
#error "VY86PID driver.c included for non-VY86PID build"
#endif /* !VY86PID */

#include        "../global.h"
#include        "driver.h"

#ifdef SMC91C92
#include        "../e9192.h"
int rdp_index = 0;

/*
 * use_serial and use_enet are used to accpet RDP communication line.
 * They are mutual exclusive.  Both are set initially to poll on the
 * comminucation channels.
 */
int use_serial = 1;
int use_enet = 1;
#endif

/*---------------------------------------------------------------------------*/

/* NOTE: This driver does not enable or disable the IRQ and FIQ
 * interrupts it merely enables the ability for an IRQ (or FIQ) to
 * occur when the IRQ (and FIQ) bits in CPSR are clear. Presumably the
 * caller of this routine has disabled both the FIQ and IRQ.
 */

const char sysdesc[] = SYSINFO ;

const char *ResetDriver(void)
{
	*((vuptr)(SER_IOBASE + IER)) = 0 ; /* disable interrupts */

	*((vuptr)(SER_IOBASE + LCR)) = DLAccess ; /* enable Divisor Latch Access */        
	*((vuptr)(SER_IOBASE + DLL)) = DLLvalue ; /* set baud rate */
	*((vuptr)(SER_IOBASE + DLM)) = DLMvalue ;
        
	/* disable Divisor Latch Access,set 9600 Baud 8 data 1 stop no parity */
	*((vuptr)(SER_IOBASE + LCR)) = LCRvalue ;
        
	/* Enable both Rx and Tx FIFOs and flush them as well */
	*((vuptr)(SER_IOBASE + FCR)) = (TxRxFIFOen | TxFIFOclear | RxFIFOclear);

#if 1 /* Do not raise RTS until we are ready for data */
	*((vuptr)(SER_IOBASE + MCR)) = (DataTransmReady | MCRInterruptEnable) ;
#else
	/* enable Modem Control Register */
	*((vuptr)(SER_IOBASE + MCR)) = MCRvalue ;
#endif

	*((vuptr)(IOBase + IRQM)) = TimerIT ; /* allow these IRQ interrupts */
	*((vuptr)(SER_IOBASE + IER))  = IERvalue ; /* enables Serial Interrupts */

#ifdef SMC91C92
	/*
	 * Initialize ethernet.
	 * -spock
	 */
	ENET_Init();
#endif

	return(&sysdesc[0]) ;
}

/*---------------------------------------------------------------------------*/
/*
 * S e t B a u d
 *
 * ENTER:       baud = magic value to set baud rate to.
 *
 * EXIT:        void
 *
 * This routine is called from DriverSpeed.
 */
void SetBaud(unsigned char baud)
{
 while ((*((vuptr)(SER_IOBASE + LSR)) & TxEmpty) == 0) ; /* wait for Transmit register to be empty */

 *((vuptr)(SER_IOBASE + IER)) = 0 ; /* disable interrupts */
 *((vuptr)(SER_IOBASE + FCR)) = 0 ; /* disable FIFO */
        
 *((vuptr)(SER_IOBASE + LCR)) |= DLAccess ; /* enable Access of Divisor Latch */
        
 *((vuptr)(SER_IOBASE + DLL)) = (baud % 256) ; /* set baud rate */
 *((vuptr)(SER_IOBASE + DLM)) = (baud / 256) ;

 *((vuptr)(SER_IOBASE + LCR)) &= ~DLAccess ; /* disable Access of Divisor Latch */

 *((vuptr)(SER_IOBASE + FCR)) = (TxRxFIFOen | TxFIFOclear | RxFIFOclear) ; /* Enable both Rx and Tx FIFOs and flush them as well*/
 *((vuptr)(SER_IOBASE + IER)) = IERvalue ; /* enables Interrupts */

 return ;
}

/*---------------------------------------------------------------------------*/
/* DriverSpeed
 * -----------
 *
 * baud : 0 = default (9600)
 *        1 = 9600
 *        2 = 19200
 *        3 = 38400
 *        4 = 57600
 *        5 = 115200
 */
void DriverSpeed(int baud)
{
 switch (baud)
  {
#if 0 /* the old system was to preserve the current rate, the new system (as specified by Stephane Thomas) explicitly sets 9600 baud */
   case 0 : /* if ZERO do NOT change current baud rate */
    break ;
#else
   case 0 : /* fall through to set 9600 */
#endif
   default : /* set baudrate to 9600 if unrecognised value */
   case 1 :
    SetBaud(BR9600) ;
    break;

   case 2 :
    SetBaud(BR19200) ;
    break;

   case 3 :
    SetBaud(BR38400) ;
    break;

   case 4 :
    SetBaud(BR57600) ;
    break;

   case 5 :
    SetBaud(BR115200) ;
    break;
  }

 return ;
}

/*---------------------------------------------------------------------------*/
/*
 * G e t B y t e
 *
 * ENTER:       void
 *
 * EXIT:        char (byte) gotten
 */
unsigned char GetByte(void)
{
	unsigned char c ;

	/* We need to ensure that the RTS line is maintained when polling
	 * data across, since the remote link may send bytes whilst we are
	 * away processing the result of a reception, causing an overflow.
	 *
	 * NOTE: The default armsd "pisd/sun/serdrive.c" driver does NOT
	 * enable the SUN CRTSCTS control. This means that the following flow
	 * control code will not affect the system. For flow control support
	 * the latest JumpStart releases must be used.
	 */
	
	*((vuptr)(SER_IOBASE + MCR)) |= (1 << 1) ; /* assert RTS */

#if 0 /* generate a softbreak instead */
	if (*((vuptr)(SER_IOBASE + LSR)) & (1 << 1)) /* overrun */
		ROMReset() ; /* a bit fatal, but maybe the best solution */
#endif

#ifdef SMC91C92
	if (!use_serial && use_enet && rdp_in) {
		goto get_netbyte;
	}

	/*
	 * If serial channel has established and we receive packets from
	 * ethernet which is a valid RDP message, we ignore the message.
	 * This is to prevent armsd trying to establish a ethernet connection
	 * while there is already one running.  We reset the byte count.
	 * The cause is that we have received an IRQ in user mode.  The IRQ routine
	 * will try to simulate a break point and return to C-Demon's control.
	 */
	if (use_serial && !use_enet && rdp_in) {
		rdp_in = 0;
		rdp_index = 0;
	}

	/* wait for character to arrive */
poll_next:

	/*
	 * No connections yet.
	 */
	if (!rdp_opened) {
		use_serial = 1;
		use_enet = 1;
	}

	while (TRUE) {
		/*
		 * Poll on serial line  receive channel.
		 */
		if (use_serial && (*((vuptr)(SER_IOBASE + LSR)) & RxDataReady)) {
			use_enet = 0;
			use_serial = 1;
			c = (unsigned char)(*((vuptr)(SER_IOBASE + RBR)));
			*((vuptr)(SER_IOBASE + MCR)) &= ~(1 << 1) ; /* drop RTS */
			return(c) ; /* return byte gotten */
		}

		/*
		 * Poll on ethernet receive channel.
		 */
		if (!use_enet)
			continue;

		*((vuptr)(ETHER_IOBASE + BankSel)) = 0x02;	/* Bank 2 */

		if (*((vuptr)(ETHER_IOBASE+IST)) & RCV_INT) {
			/*
			 * Get packet from network.
			 */
			rx_packet();
			if (rxflag == 0)
				continue;
			/*
			 * Once we have a ethernet connection, there could be a
			 * possibility that a second connection may be attempted.
			 * We ignore the second attempt.
			 */
			if (rdp_opened) {
				int i;
				for (i = 0; i < 6; i++) {
					if (Src_Addr[i] != daddr[i]) {
						rdp_in = 0;
						rdp_index = 0;
						goto poll_next;;
					}
				}
			}
			if (rdp_in == 0) {
				continue;
			}
			use_serial = 0;
			use_enet = 1;
			goto get_netbyte;
		}
	}

get_netbyte:
	/*
	 * Read a byte from RDP message from ethernet.
	 * We will enable the interrupt for the last byte.
	 */
	c = (unsigned char) *((vuptr) (RXADDR + rdp_index));
	rdp_index++;
	if ((--rdp_in) == 0) {
		rdp_index = 0;
	}
#else
	while ((*((vuptr)(SER_IOBASE + LSR)) & RxDataReady) == 0) ; /* wait for character to arrive */
	c = (unsigned char)(*((vuptr)(SER_IOBASE + RBR))) ; /* read character into c */
	*((vuptr)(SER_IOBASE + MCR)) &= ~(1 << 1) ; /* drop RTS */
#endif

	return(c) ; /* return byte gotten */
}

/*---------------------------------------------------------------------------*/
/*
 * A n y B y t e
 *
 * ENTER:       void
 *
 * EXIT:        FALSE if no char is waiting in serial Rx buffer
 *                      TRUE if a char is waiting
 */
int AnyByte(void)
{ 
#ifdef SMC91C92
	*((vuptr)(ETHER_IOBASE + BankSel)) = 0x02;	/* Bank 2 */
	if (((*((vuptr)(SER_IOBASE + LSR)) & RxDataReady) == 0) &&
	    (*((vuptr)(ETHER_IOBASE+IST)) & RCV_INT) == 0)
		return(FALSE);
#else
	if ((*((vuptr)(SER_IOBASE + LSR)) & RxDataReady) == 0)
		return(FALSE);
#endif

	return(TRUE);
}

/*---------------------------------------------------------------------------*/
/*
 * P u t B y t e
 *
 * ENTER:       c = char to send
 *
 * EXIT:        void
 */
void PutByte(unsigned char c)
{
#ifdef SMC91C92
	if (use_enet && (send_flag == 0)) {
		unsigned char xbuf[12];
		enet_header(xbuf, 1);
		xbuf[10] = c;
		ENET_TX(xbuf, 11, &daddr[0]);
	}

	if (use_serial) {
		/* wait for Transmit Holding register to be empty */
		while ((*((vuptr)(SER_IOBASE + LSR)) & TxHoldEmpty) == 0) ;

		*((vuptr)(SER_IOBASE + THR)) = c ; /* send char out serial port*/
	}
#else
	while ((*((vuptr)(SER_IOBASE + LSR)) & TxHoldEmpty) == 0) ;

	*((vuptr)(SER_IOBASE + THR)) = c ; /* send char out serial port*/
#endif

	return ;
}

/*---------------------------------------------------------------------------*/
/*
 * S e t L E D s
 *
 *-----------------------------------
 *SetLeds switches the Leds on or off
 *-----------------------------------
 *on entry : the 4 LSB of r0 contains the 4 LEDs description
 *      - Bit 0 : =0 means LED1 off;=1 means LED1 on
 *      - Bit 1 : =0   "   LED2 off;=1   "   LED2 on
 *      - Bit 2 : =0   "   LED3 off;=1   "   LED3 on
 *      - Bit 3 : =0   "   LED4 off;=1   "   LED4 on
 */
void SetLEDs(char LED)
{
 LED ^= 0xF ; /* invert last 4 bits */

#ifdef PID3_REV2
 *((vuptr)(IOBase + LEDport)) = LED;
#else
 *((vuptr)(IOBase + LEDport)) = LED << 4 ; /* on PID needs to be in bits 7 -> 4 */
#endif

 LEDimage = LED ; /* keep an image of WRITE ONLY port */
 return ;
}

/*---------------------------------------------------------------------------*/
/*
 * T i m e r I N T
 *
 * ENTER:       void
 *
 * EXIT:        void
 *
 * A timer interrupt has occurred
 */
void __irq TimerINT(void)
{
 unsigned char uc ;

 *((vuptr)(IOBase + IRQRST)) = TimerIT ; /* reset the timer compare interrupt bit */

 TickVal += 1 ; /* increment 10mS TICKER */

 if ((TickVal & 0x08) == 0x08) /* now blink the LED's at 3.125 Hz */
  uc = LEDall ; /* if off, turn ALL OFF */
 else
  uc = LEDimage ;

#ifdef PID3_REV2
 *((vuptr)(IOBase + LEDport)) = uc; /* Set LEDs to uc */
#else
 *((vuptr)(IOBase + LEDport)) = (uc << 4) ; /* Set LEDs to uc */
#endif

 return ;
}

/*---------------------------------------------------------------------------*/
/*> EOF driver.c <*/
