        TTL     C-Demon memory sizing routine                       > sizemem.s
	; "@(#)sizemem.s	1.2	96/06/07 VLSI"
        ; ---------------------------------------------------------------------
        ; This is the original assembler Demon memory sizing routine,
        ; slightly re-formatted and wrapper to allow use within the
        ; C-Demon code.
        ;
        ; $Revision: 0.1 $
        ;   $Author: jsmith $
        ;     $Date: 941101 $
        ;
        ; Original copyright (c) 1992, Advanced RISC Machines Limited.
        ; All Rights Reserved.
        ; ---------------------------------------------------------------------

        AREA    |C-Demon$$SizeMem|,CODE,READONLY

;/****************************************************************************
;* checkmem.s                                                                *
;* Version 1.00                                                              *
;* Non-disruptive memory checker and sizer.                                  *
;* Copyright (C) 1992 Advanced RISC Machines Limited. All rights reserved.   *
;* Written by Dave Flynn.                                                    *
;* Project started on 2nd January 1992.                                      *
;* Last change was 2nd January 1992.                                         *
;\****************************************************************************

        ; ---------------------------------------------------------------
        ; Find size of main RAM and return in r0. r1-r7 corrupted.

        EXPORT  FindRAMSize
FindRAMSize
        STMFD   sp!,{r1-r7}     ; store work registers

        MOV     r0,#0x00000000  ; start at base of physical RAM

        ; In reality most memory systems are some power-of-2 in size.
        ; This means that you do not need to check every byte/word in
        ; memory when sizing, only the powers.
        MOV     r1,#(8 << 10)   ; size in 8K chunks for the moment


        ; Non-invasive memory sizer
        ;  parameters:  r0 is base pointer
        ;               r1 is test increment (multiple of 4 please)
        ;  return:      r0 is size (bytes)
        ;  registers    r1-r7 corrupted.
        ;
        ; Memory is left intact after the size test.
        ;  register usage:
        ;                  +------+
        ;   base(r6)  ---->|(base)| (save old -> r4)
        ;       inc in r7  |      |
        ;                  |      |
        ;   probe(r5) ---->|      | (save in r1)
        ;                  +------+
        ;   wrapped image  |(base)|
        ; basic algorithm:
        ;      save copy of base location
        ;      write pattern to base
        ; loop save copy of word from memory
        ;      write a pattern, read back and exit if write failed
        ;      write inverted pattern, read back and exit as above
        ;      read base and see if probe has wrapped
        ;      restore probe location
        ;      advance probe and loop if base still okay
checksize
        MOV     r7,r1                   ; keep step size in r7
        MOV     r6,r0                   ; keep base address in r6
        MOV     r5,r6                   ; r5 is probe address
        LDR     r4,[r6,#0x00]           ; preserve original base contents
        MOV     r2,#0x55                ; build adjacent bit tests
        ORR     r2,r2,#0xAA00           ; make AA55 pattern
        ORR     r2,r2,r2,LSL #16        ; final AA55AA55 (helps check endian-ness)
        MVN     r3,r2                   ; use inverse copy as alternate pattern

        ; Move up memory in steps, probing each location with two different
        ; patterns to see if memory is present, and checking the base of
        ; memory too (to detect aliasing of wrapped memory).

        ; save AA55AA55 in base which should be there until wrapped
        STR     r2,[r6,#0x00]   ; store first pattern at base
        LDR     r0,[r6,#0x00]   ; and read back
        CMP     r0,r2           ; check okay
        BNE     checksizefail   ; give up if not

        ; byte testing valuable as ARM broadcasts 4 bytes
        ; and any faulty write enables will corrupt other bytes

checkbyte0
        ; check byte writes work independently (endian-independent)

        LDRB    r1,[r6,#0x00]   ; byte at a time
        MOV     r0,#0xA5        ; using local pattern
        STRB    r0,[r6,#0x00]   ; test byte 0
        LDRB    r0,[r6,#0x00]   ; read back
        CMP     r0,#0xA5        ; test
        BNE     checksizefail   ; fail if different
        STRB    r1,[r6,#0x00]   ; restore byte
        LDR     r0,[r6,#0x00]   ; test word still intact
        CMP     r0,r2           ; still pattern one
        BNE     checksizefail   ; fail if different

        ; in line for zero stack usage, etc.

checkbyte1
        LDRB    r1,[r6,#0x01]   ; preserve byte 1
        MOV     r0,#0xA5        ; using local pattern
        STRB    r0,[r6,#0x01]   ; test byte 1
        LDRB    r0,[r6,#0x01]   ; read back
        CMP     r0,#0xA5        ; test
        BNE     checksizefail   ; fail if different
        STRB    r1,[r6,#0x01]   ; restore byte
        LDR     r0,[r6,#0x00]   ; test word still intact
        CMP     r0,r2           ; still pattern one
        BNE     checksizefail   ; fail if different
checkbyte2
        LDRB    r1,[r6,#0x02]   ; preserve byte 2
        MOV     r0,#0xA5        ; using local pattern
        STRB    r0,[r6,#0x02]   ; test byte 2
        LDRB    r0,[r6,#0x02]   ; read back
        CMP     r0,#0xA5        ; test
        BNE     checksizefail   ; fail if different
        STRB    r1,[r6,#0x02]   ; restore byte
        LDR     r0,[r6,#0x00]   ; test word still intact
        CMP     r0,r2           ; still pattern one
        BNE     checksizefail   ; fail if different
checkbyte3
        LDRB    r1,[r6,#0x03]   ; preserve byte 3
        MOV     r0,#0xA5        ; using local pattern
        STRB    r0,[r6,#0x03]   ; test byte 3
        LDRB    r0,[r6,#0x03]   ; read back
        CMP     r0,#0xA5        ; test
        BNE     checksizefail   ; fail if different
        STRB    r1,[r6,#0x03]   ; restore byte
        LDR     r0,[r6,#0x00]   ; test word still intact
        CMP     r0,r2           ; still pattern one
        BNE     checksizefail   ; fail if different

        ADD     r5,r6,r7        ; r5 is probe address (base+step)
checksizeloop
        LDR     r1,[r5,#0x00]   ; preserve probed value in r1
        STR     r2,[r5,#0x00]   ; store first pattern
        LDR     r0,[r5,#0x00]   ; read it back
        CMP     r0,r2           ; is it correct?
        BNE     checksizefail   ; no working memory here
        STR     r3,[r5,#0x00]   ; store second pattern
        LDR     r0,[r5,#0x00]   ; read it back
        CMP     r0,r3           ; is it correct?
        BNE     checksizefail   ; not working

        ; if reads and writes okay did last write wrap to base?
        LDR     r0,[r6,#0x00]
        STR     r1,[r5,#0x00]   ; restore probed value
        CMP     r0,r2           ; base still reads as first pattern?
        ADDEQ   r5,r5,r7        ; if okay then increment probe
        BEQ     checksizeloop   ; and loop, else fall through to fail

        ; Address of first failing location in r2, original contents in r6
        ; Original contents of base in r7

checksizefail
        STR     r1,[r5,#0x00]   ; restore original contents of probe
        STR     r4,[r6,#0x00]   ; restore original contents of base
        SUB     r0,r5,r6        ; valid size is (probe - base)

        LDMFD   sp!,{r1-r7}     ; recover registers corrupted by the checksize code
        MOV     pc,lr

        ; ---------------------------------------------------------------------
        END     ; EOF sizemem.s
