/*> swi.c <*/
/* C-Demon SWI processing.
 *
 * $Revision: 0.1 $
 *   $Author: jsmith $
 *     $Date: 941031 $
 *
 * Copyright (c) 1994, VLSI Technology Inc. All Rights Reserved.
 *
 * "@(#)swi.c	1.2	96/06/07 VLSI"
 */
/*---------------------------------------------------------------------------*/

#include    "global.h"
#include    "driver.h"
#include    "bp.h"
#include    "rdi.h"
#include    "swi.h"

/*---------------------------------------------------------------------------*/
/*-- Internal SWI support manifests -----------------------------------------*/
/*---------------------------------------------------------------------------*/
/* Following generated on 2 bit basis where:
 *
 *    N = 00   None
 *    B = 01   Byte
 *    W = 10   Word
 *    S = 11   String
 *  BLK = 1110 (looks like SW but is a SPECIAL code)
 */
#define RQ_N        (0x00)  /* No byte, word, or string */
#define RQ_B        (0x01)  /* single byte */
#define RQ_W        (0x02)  /* single word */
#define RQ_S        (0x03)  /* simple string */
#define RQ_BLK      (0x0E)  /* Actually sent as SW but not really a string (SWI_Write) */
#define RQ_BS       (0x07)  /* string then byte (SWI_Open) */
#define RQ_SS       (0x0F)  /* string then another string (SWI_Rename) */
#define RQ_WW       (0x0A)  /* word then a word (SWI_Seek and SWI_IsTTY) */
#define RQ_WWW      (0x2A)  /* word then a word and another word (SWI_Read) */

/*---------------------------------------------------------------------------*/
/*-- Internal SWI support routines ------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* StrLen
 * in:  r0 = location of NUL terminated ASCII string
 * out: r0 = length of string (excluding terminating NUL)
 */
static int StrLen(char *cp)
{
 int len = 0 ;
 while (*cp++ != '\0') /* NUL */
  len++ ;
 return(len) ;
}

/*---------------------------------------------------------------------------*/
/* SendBlock
 * =========
 *
 * ENTER:   cp = loc of block
 *          len = len of block
 *
 * EXIT:    void
 *
 * This routine has three algorythms.
 *
 * 1. If the len is <= 32 then the len and the bytes of block are sent to HOST.
 * 2. if the len is >32 and <255 then the len and the loc is sent to HOST.
 * 3. if the len is >= 255 then 255 followed by len and then loc is sent to HOST.
 */
static void SendBlock(int len,char *cp)
{
 int j ;
 
 if (len >= 255)
  {
   PutByte(255) ;       /* tell HOST this is not count,it is in the following word */
   PutWord(len) ;       /* send the length */
   PutWord((int)cp) ;   /* and the location */
  }
 else
  {
   PutByte(len) ;       /* send block len to the HOST */
   if (len > 32)
    PutWord((int)cp) ;  /* and the location */
   else
    {
     if (len > 0)
      {
       for (j=0; (j < len); j++)
        PutByte(*cp++) ; /* send byte(s) to HOST */
      }
    }
  }
 return ;
}

/*---------------------------------------------------------------------------*/
/* SendStr
 * =======
 *
 * ENTER:       cp = loc of string
 *
 * EXIT:        void
 *
 * Send a NULL terminated string pointed at by r0 to HOST
 */
static void SendStr(char *cp)
{
 SendBlock(StrLen(cp),cp) ; /* send the block to the HOST */
 return ;
}

/*---------------------------------------------------------------------------*/
/* SWI_Send
 * ========
 *
 * ENTER:   typ = Type of Send
 *          num = SWI number (OS Request OP)
 *
 * This routine works closely with "armos.c" which provides the system interface
 * on the HOST side. The other SWI functions use this as a rear-end and branch
 * here after setting the type of parameters passed in the r8 register. If a
 * ZERO is passed it indicates that only the SWI number in r7 is to be sent to
 * the HOST. If a single byte or word is indicated it is passed in the r0
 * register. Further bytes (and/or words) are passed in r1, then r2, and finally
 * in r3. If a string is passed then the respective register contains a pointer.
 * At most only 4 simultaneous requests can be made in a single call. This is
 * because the OS Request OP is but a byte and the bits are organized as pairs
 * as 44332211 where any pair has the following decode:
 *
 *  xx | FUNCTION
 *  ---+---------
 *  00 | None (END or respond only to SWI number)
 *  01 | byte follows
 *  10 | word follows
 *  11 | string follows
 *
 * For exceptions to these simple rules see armos.c
 *
 * This version of send uses the fact that all functions that use send pass one
 * of the following:
 *
 *  TYPE    DESCRIPTION
 *  RQ_N    No byte, word, or string
 *  RQ_B    single byte
 *  RQ_W    single word
 *  RQ_S    simple string
 *  RQ_BLK  Actually coded as RQ_SW (SWI_Write)
 *  RQ_BS   string then byte (SWI_Open)
 *  RQ_SS   string then another string (SWI_Rename)
 *  RQ_WW   word then a word (SWI_Seek and SWI_IsTTY)
 *  RQ_WWW  word then a word and another word (SWI_Read)
 *
 * NOTE: Due to self imposed limitations a RQ_SW cannot be used as it has been
 *  used for the RQ_BLK.
 */
static void SWI_Send(int num,unsigned char typ)
{
 PutByte(RDP_OSOp) ; /* Generate an OS Request */
 PutWord(num) ;      /* send the SWI we were invoked by */
 PutByte(typ) ;      /* then the code of what else is coming */
 return ;
}

/*---------------------------------------------------------------------------*/
/* SWI_Sent
 * ========
 *
 * ENTER:   void
 *
 * EXIT:    int
 *
 * Common completion code for SWI_Send functions.
 *
 * NOTE: The use of "sts" below is a bodge. The OSOp replies are
 * placed directly into the processor state by the host debugger. The
 * Level1RDI function is called to process the messages from the host
 * debugger, until the OSOpReply has been seen. At this point the code
 * can continue. The "sts" variable is used to ensure that "a1" is
 * returned to the caller of SWI_Sent. The Level1RDI processing will
 * have written r0/a1 depending on the OSOp message given. This is
 * nasty, since it is relying on the compiler not generating code that
 * uses r0 between completion of the Level1RDI call, and falling out
 * of the SWI_Sent code.
 */
/* This is a nasty bodge. The Level1RDI function is actually a void
 * routine. However this ensures that the r0/a1 value that results
 * from Level1RDI can be passed through to the caller of SWI_Sent.
 */
extern int Level1RDI(int num) ;

static int SWI_Sent(void)
{
 unsigned char c ;
 int sts = 0x00 ;

 do
  {
   c = GetByte() ;               /* get byte from HOST */
#if 1 /* ensure compiler leaves r0/a1 alone from RDP processing */
   sts = Level1RDI(c) ;          /* handle response through RDI */
#else
   (void)Level1RDI(c) ;          /* handle response through RDI */
#endif
  } while (c != RDP_OSOpReply) ; /* until host replies to the SWI */

 return(sts) ;
}

/*---------------------------------------------------------------------------*/
/*-- Exported SWI support routines ------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* SWI_init
 * ========
 *
 * This function must be called prior to using any SWI calls so that initial
 * conditions are properly set.
 */
void SWI_init(void)
{
 /* Set default of OK (normal exit) */
 BP_RDP_STS = RDIError_NoError ;
 BP_CPSR = 0x10 ; /* set USR32 mode as default (nzcvift_USR32) */
 return ;
}

/*---------------------------------------------------------------------------*/
/*-- locally handled SWI functions ------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* inSWIExit
 * =========
 *
 * ENTER:   void
 *
 * EXIT:    void
 *
 * This code is entered via an SWI Exit or an SWI Break
 */
void inSWIExit(void)
{
 if ((DebuggeeRunning & 0x1) == RDP_ExecSync) /* check bit 0 only */
  PutByte(RDP_Return) ;
 else
  PutByte(RDP_Stopped) ;

 if (DebuggeeRunning & (1 << 7)) /* check for PointHandle return */
  PutWord(0xFFFFFFFF) ; /* send undefined point handle */
 
 DebuggeeRunning = 0 ; /* reset it to NOT running state */
 PutByte(BP_RDP_STS) ; /* send reason for EXIT */
    
 /* Reset reason for exit to default */
 BP_RDP_STS = RDIError_NoError ;

 return ;
}

/*---------------------------------------------------------------------------*/
/* inSWIClock
 * ==========
 *
 * ENTER:   void
 *
 * EXIT:    time in centi-seconds ( 10 mS ticks)
 *
 * Return, in r0, the number of centi-seconds since the support code began 
 * execution. In general, only the difference between successive calls to 
 * <SWI_Clock>, can be meaningful.
 */
int inSWIClock(void)
{
 return(ReadTimer() - TimeZero) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIClockInit
 * ==============
 *
 * ENTER:   void
 *
 * EXIT:    void
 *
 * Must be invoked prior to invoking SWIClock
 */
void inSWIClockInit(void)
{
 TimeZero = ReadTimer() ;
 return ;
}

/*---------------------------------------------------------------------------*/
/* inSWIInstallHandler
 * ===================
 *
 * ENTER:   vecnum = Exception Number (from table below)
 *          vecval = Value for new handler when invoked
 *          vecp   = pointer to new handler
 *
 * EXIT:    p1 = vecnum
 *          p2 = private argument for previous handler
 *          p3 = pointer to previous handler
 *
 * <SWI_InstallHandler> installs a handler for a hardware
 * exception. On entry, r0 contains the exception number (see the
 * table below); r1 contains a value to pass to the handler when it is
 * eventually invoked; and r2 contains the address of the handler. On
 * return, r2 contains the address of the previous handler and r1
 * contains its argument.
 *
 * NOTE: This is NOT completely true for the C-Demon world.
 *
 * On occurrence of the exception, the handler is entered in the
 * appropriate non-user mode, with r10 holding a value dependent on
 * the exception type, and r11 holding the handler argument (as passed
 * to InstallHandler in r1). r10,r11,r12 and r14 (for the processor
 * mode in which the handler is entered) are saved on the stack of the
 * mode in which the handler is entered; all other registers are as at
 * the time of the exception. (Any effects of the instruction causing
 * the exception have been unwound). If the handler returns, the
 * exception will be passed to the debugger (regardless of the value
 * of the debugger variable $vector_catch).
 *
 * The C-Demon vector numbers. NOTE: The IRQ and FIQ vectors by bit
 * depend on the target architecture.
 *
 * VECNUM EXCEPTION               MODE      r10 VALUE
 * ---------------------------------------------------
 * 0x10-0x1F  FIQ vector by bit      FIQ32  -
 * 0x20-0x3F  IRQ vector by bit      IRQ32  -
 * 0x40       branch through zero    SVC32  -
 * 0x41       undefined instruction  UND32  -
 * 0x42       swi                    SVC32  swi number
 * 0x43       prefetch abort         ABT32  -
 * 0x44       data abort             ABT32  -
 * 0x45       address exception      SVC32  -
 * 0x46       IRQ                    IRQ32  -
 * 0x47       FIQ                    FIQ32  -
 * 0x48       Error                  SVC32  error pointer
 * 0x49       Get Byte from HOST     -      -
 * 0x4A       Put Byte to HOST       -      -
 * 0x4B       Get WORD from HOST     -      -
 * 0x4C       Put WORD to HOST       -      -   
 * 0x4D Print ASCIIZ String to HOST  -      -
 * 0x4E       Set LEDs               -      -
 *
 * The reason the C-Demon vector numbers are different from the
 * original assembler Demon is to avoid the problem with the run-time
 * support library updating the vectors, and then not being able to
 * cope with the resulting interrupts and events. This version of
 * SWI_InstallHandler will ignore vectors it does not know about.
 */
__value_in_regs ret3parm inSWIInstallHandler(int vecnum,int vecval,unsigned int *vecp)
{
 ret3parm retval ;
 unsigned int *uip ;

 /* We return the vector number in "p1", or 0x00000000 if the vector
  * number given was invalid.
  */
 retval.p1 = vecnum ; /* return vecnum in p1 */

 /* We ignore vectors that are not in the correct ranges */
 switch (vecnum & 0xF0) /* check main range first */
  {
   case 0x10 : /* 16 possible FIQ bit vectors */
    if ((vecnum & 0x0F) >= NUMFIQS)
     uip = NULL ;
    else
     uip = (&FIQvectors + (vecnum & 0x0F)) ;
    break;

   case 0x20 : /* 32 possible IRQ bit vectors */
   case 0x30 :
    if ((vecnum & 0x1F) >= NUMIRQS)
     uip = NULL ;
    else
     uip = (&IRQvectors + (vecnum & 0x1F)) ;
    break;

   case 0x40 : /* C-Demon system vectors */
    {
     switch (vecnum)
      {
       case 0x40 : /* branch through zero */
        uip = (&ResetV) ;
        break;

       case 0x41 : /* undefined instruction */
        uip = (&UndefInstrV) ;
        break;

       case 0x42 : /* SWI */
        uip = (&SWIV) ;
        break;

       case 0x43 : /* prefetch abort */
        uip = (&PrefAbortV) ;
        break;

       case 0x44 : /* data abort */
        uip = (&DataAbortV) ;
        break;

       case 0x45 : /* address exception */
        uip = (&AddrExceptV) ;
        break;

       case 0x46 : /* IRQ */
        uip = (&IRQV) ;
        break;

       case 0x47 : /* FIQ */
        uip = (&FIQV) ;
        break;

       case 0x48 : /* Error */
        uip = (&ErrorV) ;
        break;

       case 0x49 : /* Get Byte */
        uip = (&GetByteV) ;
        break;

       case 0x4A : /* Put Byte */
        uip = (&PutByteV) ;
        break;

       case 0x4B : /* Get Word */
        uip = (&GetWordV) ;
        break;

       case 0x4C : /* Put Word */
        uip = (&PutWordV) ;
        break;

       case 0x4D : /* Print ASCIIZ string to HOST */
        uip = (&PrintStringV) ;
        break;

       default : /* unsupported handler */
        uip = NULL;
        break;
      }
    }
    break ;

   default : /* unrecognised vector */
    uip = NULL ;
    break ;
  }

 /* The armlib "__osdep_traphandlers_init" code expects the vector
  * number to remain untouched by the routine. This means we cannot
  * return an error status easily without also updating the standard
  * ARM library run-time for Demon systems.
  */
 retval.p2 = 0xDEADC0DE ; /* This is not used at the moment */
 if (uip == NULL)
  retval.p3 = NULL ; /* no previous handler, since vector not supported */
 else
  {
   retval.p3 = *uip ; /* pointer to previous handler */
   *uip = (unsigned int)vecp ; /* set new handler */
  }

 return(retval) ; /* pass the structure back */
}

/*---------------------------------------------------------------------------*/
/* NOTE: This is a C-Demon extension (over the basic Demon system) */
/* inSWIGetVector
 * ==============
 *
 * ENTER:   vecnum = Exception Number (from table below)
 *
 * EXIT:    p1 = NULL then Vector does NOT exist, else handler slot location.
 *
 * SWIGetVector uses the same vecnum as SWIInstallHandler, but only returns
 * the location of the current vector, not its contents
 *
 * VECNUM EXCEPTION               
 * -----------------------------
 * 0x10-0x17  FIQ vector by bit
 * 0x20-0x3F  IRQ vector by bit
 * 0x40       branch through zero
 * 0x41       undefined instruction
 * 0x42       swi
 * 0x43       prefetch abort
 * 0x44       data abort
 * 0x45       address exception
 * 0x46       IRQ
 * 0x47       FIQ
 * 0x48       Error
 * 0x49       Get Byte from HOST
 * 0x4A       Put Byte to HOST
 * 0x4B       Get WORD from HOST
 * 0x4C       Put WORD to HOST  
 * 0x4D       Print ASCIIZ String to HOST
 * 0x4E       Set LEDs
 */

void *inSWIGetVector(int vecnum)
{
 unsigned int *uip ;

 switch (vecnum & 0xF0) /* main vector range */
  {
   case 0x10 : /* 16 possible FIQ bit vectors */
    if ((vecnum & 0x0F) >= NUMFIQS)
     uip = NULL ;
    else
     uip = (&FIQvectors + (vecnum & 0x0F)) ;
    break;

   case 0x20 : /* 32 possible IRQ bit vectors */
   case 0x30 :
    if ((vecnum & 0x1F) >= NUMIRQS)
     uip = NULL ;
    else
     uip = (&IRQvectors + (vecnum & 0x1F)) ;
    break ;

   case 0x40 :
    switch (vecnum)
     {
      case 0x40 : /* branch through zero */
       uip = (&ResetV) ;
       break ;

      case 0x41 : /* undefined instruction */
       uip = (&UndefInstrV) ;
       break ;

      case 0x42 : /* SWI */
       uip = (&SWIV) ;
       break ;

      case 0x43 : /* prefetch abort */
       uip = (&PrefAbortV) ;
       break ;

      case 0x44 : /* data abort */
       uip = (&DataAbortV) ;
       break ;

      case 0x45 : /* address exception */
       uip = (&AddrExceptV) ;
       break ;

      case 0x46 : /* IRQ */
       uip = (&IRQV) ;
       break ;

      case 0x47 : /* FIQ */
       uip = (&FIQV) ;
       break ;

      case 0x48 : /* Error */
       uip = (&ErrorV) ;
       break ;

      case 0x49 : /* Get Byte */
       uip = (&GetByteV) ;
       break ;

      case 0x4A : /* Put Byte */
       uip = (&PutByteV) ;
       break ;

      case 0x4B : /* Get Word */
       uip = (&GetWordV) ;
       break ;

      case 0x4C : /* Put Word */
       uip = (&PutWordV) ;
       break ;

      case 0x4D : /* Print ASCIIZ string to HOST */
       uip = (&PrintStringV) ;
       break ;

      default : /* unsupported handler */
       uip = NULL ;
       break ;
     }
    break ;

   default : /* unsupported handler */
    uip = NULL ;
    break ;
  }

 return((void *)uip) ; /* return location of handler */
}

/*---------------------------------------------------------------------------*/
/* inSWIGenerateError
 * ==================
 *
 * ENTER:   ep = pointer to ERROR BLOCK
 *          rp = pointer to REGISTER BLOCK
 *
 * EXIT:    void
 *
 * On entry, r0 points to an error block (containing a 32-bit error number, 
 * followed by a zero-terminated error string); r1 points to a 17-word block 
 * containing the values of the ARM CPU registers at the instant the error 
 * occurred (the 17th word contains the PSR).
 *
 * <SWI_GenerateError> causes the (software) error vector to be called: if bit 8 
 * of $vector_catch is set this causes the debugger to be entered directly; 
 * otherwise, the installed error handler is called (see <SWI_InstallHandler>).
 */
void inSWIGenerateError(int *ep,int *rp)
{
 /* This is not valid. The "ep" and "rp" parameters need to be passed
  * through to the called routine:
  */
 FuncIndirectZP((unsigned int *)&ErrorV) ;
 return ; /* might never be reached */
}

/*---------------------------------------------------------------------------*/
/*-- host supported SWI functions -------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* inSWIWriteC
 * ===========
 *
 * ENTER:   ub  = byte to send to HOST
 *          num = 0x00
 *
 * EXIT:    void
 *
 * Write a byte, passed in register 0, to the debug channel. When executed under 
 * <armsd>, the character will appear on the display device connected to <armsd>.
 */
void inSWIWriteC(unsigned char ub,int num)
{
 SWI_Send(num,RQ_B) ; /* sending byte to HOST */
 PutByte(ub) ;        /* send the byte */
 (void)SWI_Sent() ;
 return ;
}

/*---------------------------------------------------------------------------*/
/* inSWIWrite0
 * ===========
 *
 * ENTER:   num = 0x02
 *          cp  = pointer to NULL terminated string
 *
 * EXIT:    void
 *
 * Write the null terminated string, pointed to by register 0, to the debug 
 * channel. When executed under <armsd>, the characters will appear on the
 * display device connected to <armsd>.
 */
void inSWIWrite0(char *cp,int num)
{
 SWI_Send(num,RQ_S) ; /* sending a string to HOST */
 SendStr(cp) ;        /* send the string */
 (void)SWI_Sent() ;
 return ;
}

/*---------------------------------------------------------------------------*/
/* inSWINewLine
 * ============
 *
 * ENTER:   num = 0x03
 *
 * EXIT:    void
 *
 */
void inSWINewLine(int num)
{
 SWI_Send(num,RQ_N) ; /* have HOST generate a New Line */
 (void)SWI_Sent() ;
 return ;
}

/*---------------------------------------------------------------------------*/
/* inSWIReadC
 * ==========
 *
 * ENTER:   num = 0x04
 *
 * EXIT:    return a byte from HOST
 *
 * Read a byte from the debug channel, returning it in register 0. The read is 
 * notionally from the keyboard attached to the debugger.
 */
unsigned char inSWIReadC(int num)
{
 SWI_Send(num,RQ_N) ; /* get a byte from the HOST */
 return((unsigned char)SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWICLI
 * ========
 *
 * ENTER:   num = 0x05
 *          cp  = pointer to string for HOST command line
 *
 * EXIT:    void
 *
 * Pass the string, pointed to by register 0, to the host's command line 
 * interpreter (NOT AVAILABLE IN THE PC/DOS RELEASE).
 */
void inSWICLI(char *cp,int num)
{
 SWI_Send(num,RQ_S) ; /* sending a string to HOST */
 SendStr(cp) ;        /* send the string */
 (void)SWI_Sent() ;
 return ;
}

/*---------------------------------------------------------------------------*/
/* inSWIGetEnv
 * ===========
 *
 * ENTER:   num = 0x10
 *
 * EXIT:    r0 = pointer to command line used to invoke program
 *          r1 = size of RAM
 *
 * Returns in register 0 the address of the command line string used to invoke
 * the program, and in register 1 the highest available address in user memory.
 */
__value_in_regs ret2parm inSWIGetEnv(int num)
{
 ret2parm retval ;
 char *cp ;
 
 cp = (char *)&CommandLine ;    /* point cp at loc of CommandLine */
 SWI_Send(num,RQ_W) ;
 PutWord((int)cp) ;             /* send it to HOST */
 (void)SWI_Sent() ;

 retval.p1 = (unsigned int)&CommandLine ;
 retval.p2 = RAMSize ;
    
 return(retval) ;
}

/*---------------------------------------------------------------------------*/
/* inSWITime
 * =========
 *
 * ENTER:   num = 0x63
 *
 * EXIT:    current time as elapsed seconds from 1970 JAN 01 12:00 am
 *
 * Return, in r0, the number of seconds since the beginning of 1970 (the Unix
 * time origin).
 */
int inSWITime(int num)
{
 SWI_Send(num,RQ_N) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIRemove
 * ===========
 *
 * ENTER:   num = 0x64
 *          cp  = pointer to file name as NULL terminated string
 *
 * EXIT:    0 if remove is successful
 *          non-ZERO error code (host specific) otherwise
 *
 * Delete (remove, un-link, wipe, destroy) the file named by the NUL-terminated 
 * string addressed by r0. Return (in r0) a <zero> if the removal succeeds, or a 
 * <non-zero>, host-specific error code if it fails.
 */
int inSWIRemove(char *cp,int num)
{
 SWI_Send(num,RQ_S) ;
 SendStr(cp) ;
 return(SWI_Sent()) ; /* send string pointed at with r0 to HOST */
}

/*---------------------------------------------------------------------------*/
/* inSWIRename
 * ===========
 *
 * ENTER:   num = 0x65
 *          cpo = Pointer to original file name as NULL terminated string
 *          cpn = Pointer to desired file name as NULL terminated string
 *
 * EXIT:    0 then rename was successful
 *          non-ZERO is error code (host specific)
 *
 * R0 and r1 address NUL-terminated strings, the <old-name> and <new-name>, 
 * respectively. If the rename succeeds, zero is returned in r0; otherwise, a 
 * non-zero, host-specific error code is returned.
 */
int inSWIRename(char *cpo,char *cpn,int num)
{
 SWI_Send(num,RQ_SS) ; /* sending two strings to HOST */
 SendStr(cpo) ;
 SendStr(cpn) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIOpen
 * =========
 *
 * ENTER:   num = 0x66
 *
 * EXIT:    0 then the open has failed
 *          non-ZERO handle value and open was successful
 *
 * R0 addresses a NUL-terminated string containing a file or device name; r1 is a 
 * small integer specifying the file opening mode (see table below). If the open 
 * succeeds, a non-zero <handle> is returned in r0, which can be quoted to 
 * <SWI_Close>, <SWI_Read>, <SWI_Write>, <SWI_Seek>, <SWI_Flen> and <SWI_IsTTY>. 
 * Nothing else may be asserted about the value of the <handle>. If the open 
 * fails, the value 0 is returned in r0.
 *
 *    r1 value      ANSI C fopen() mode
 *
 *    0             "r"
 *    1             "rb"
 *    2             "r+"
 *    3             "r+b"
 *
 *    4             "w"
 *    5             "wb"
 *    6             "w+"
 *    7             "w+b"
 *
 *    8             "a"
 *    9             "ab"
 *    10            "a+"
 *    11            "a+b"
 */
int inSWIOpen(char *cp,unsigned char ub,int num)
{
 SWI_Send(num,RQ_BS) ; /* sending string and a byte to HOST */
 SendStr(cp) ;
 PutByte(ub) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIClose
 * ==========
 *
 * ENTER:   num    = 0x68
 *          handle = handle (from OPEN or RENAME)
 *
 * EXIT:    0 if close is successful
 *          non-ZERO error code (host specific)
 *
 * On entry, r0 must be a <handle> for an open file, previously returned by 
 * <SWI_Open>. If the close succeeds, zero is returned in r0; otherwise, a 
 * non-zero value is returned.
 */
int inSWIClose(int handle,int num)
{
 SWI_Send(num,RQ_W) ; /* sending word to HOST */
 PutWord(handle) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIWrite
 * ==========
 *
 * ENTER:   num    = 0x69
 *          handle = handle (from OPEN or RENAME)
 *          cp     = pointer to buffer containing what to write
 *          len    = number of bytes to write
 *
 * EXIT:    0 then write was successful
 *          non-ZERO as number of bytes NOT written
 *
 * On entry, r0 must contain a handle for a previously opened file; r1 points to
 * a buffer in the callee; and r2 contains the number of bytes to be written from
 * the buffer to the file. SWI_Write returns, in r0, the number of bytes <not>
 * written (and so indicates success with a 0 return value).
 */
int inSWIWrite(int handle,char *cp,int len,int num)
{
 SWI_Send(num,RQ_BLK) ; /* sending BLOCK to HOST */
 PutWord(handle) ;
 SendBlock(len,cp) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIRead
 * =========
 *
 * ENTER:   num    = 0x6A
 *          handle = handle (from OPEN or RENAME)
 *          cp     = buffer to fill with read bytes
 *          len    = number of bytes to read
 *
 * EXIT:    0 then read was successful
 *          non-ZERO as number of bytes NOT read
 *
 * On entry, r0 must contain a handle for a previously opened file or device; r1 
 * points to a buffer in the callee; and r2 contains the number of bytes to be 
 * read from the file into the buffer. <SWI_Read> returns, in r0, the number of 
 * bytes <not> read, and so indicates the success of a read from a file with a 
 * zero return value. If the handle is for an interactive device (<SWI_IsTTY> 
 * returns non-zero for this handle), then a non-zero return from <SWI_Read> 
 * indicates that the line read did not fill the buffer.
 */
int inSWIRead(int handle,char *cp,int len,int num)
{
 SWI_Send(num,RQ_WWW) ; /* sending info (three words) to HOST */
 PutWord(handle) ;
 PutWord((int)cp) ;
 PutWord(len) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWISeek
 * =========
 *
 * ENTER:   num    = 0x6B
 *          handle = handle (from OPEN or RENAME)
 *          pos    = byte position in file to position to
 *
 * EXIT:    0 then seek was successful
 *          non-ZERO error code (host specific)
 *
 * On entry, r0 must contain a handle for a seekable file object, and r1 the 
 * absolute byte position to be sought to. If the request can be honoured then 
 * <SWI_Seek> returns 0 in r0; otherwise it returns a host-specific non-zero 
 * value. Note that the effect of seeking outside of the current extent of the 
 * file object is undefined.
 */
int inSWISeek(int handle,int pos,int num)
{
 SWI_Send(num,RQ_WW) ; /* sending info (two words) to HOST */
 PutWord(handle) ;
 PutWord(pos) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIFlen
 * =========
 *
 * ENTER:   num = 0x6C
 *          handle = handle (from OPEN or RENAME)
 *
 * EXIT:    -1 if unable to return current length of file
 *          otherwise it is the length of file
 *
 * On entry, r0 contains a handle for a previously opened, seekable file object. 
 * <SWI_Flen> returns, in r0, the current length of the file object, otherwise it 
 * returns -1.
 */
int inSWIFlen(int handle,int num)
{
 SWI_Send(num,RQ_W) ; /* sending info (one word) to HOST */
 PutWord(handle) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIIsTTY
 * ==========
 *
 * ENTER:   num    = 0x6E
 *          handle = handle (from OPEN or RENAME)
 *
 * EXIT:    1 if device is TTY (or CONSOLE or interactive)
 *          0 device is NOT a TTY (or CONSOLE or interactive)
 *
 * On entry, r0 must contain a handle for a previously opened file or device 
 * object. On exit, r0 contains 1 if the handle identifies an interactive device;
 * otherwise r0 contains 0.
 */
int inSWIIsTTY(int handle,int num)
{
 SWI_Send(num,RQ_W) ; /* sending info (one word) to HOST */
 PutWord(handle) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWITmpNam
 * ===========
 *
 * ENTER:   num = 0x6F
 *          cp  = pointer to buffer large enough to contain L_tmpnam bytes
 *          len = length of buffer (len >= L_tmpnam )
 *
 * EXIT:    0 then request for temp name FAILED
 *          non-ZERO temp name is in buffer
 *
 * On entry, r0 points to a buffer and r1 contains the length of the buffer (r1 
 * should be at least the value of L_tmpnam on the host system). On successful 
 * return, r0 points to the buffer which contains a host temporary file name. If 
 * the request cannot be satisfied (e.g. because the buffer is too small) then 0 
 * is returned in r0.
 */
int inSWITmpNam(char *cp,int len,int num)
{
 SWI_Send(num,RQ_WW) ; /* sending info (two words) to HOST */
 PutWord((int)cp) ;
 PutWord(len) ;
 return(SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/* inSWIGetErrno
 * =============
 *
 * ENTER:   num = 0x60
 *
 * EXIT:    ERRNO from C lib
 *
 * Return, in r0, the value of the C library <errno> variable associated with the
 * host support for this debug monitor. <errno> may be set by a number of C 
 * library support SWIs, including <SWI_Remove>, <SWI_Rename>, <SWI_Open>, 
 * <SWI_Close>, <SWI_Read>, <SWI_Write>, <SWI_Seek>, etc. Whether or not, and to 
 * what value <errno> is set is completely host-specific, except where the ANSI C
 * standard defines the behaviour.
 */
int inSWIGetErrno(int num)
{
 SWI_Send(num,RQ_N) ; /* get errno from the HOST */
 return((unsigned char)SWI_Sent()) ;
}

/*---------------------------------------------------------------------------*/
/*> swi.c <*/
